(* Time-stamp: <modified the 09/10/2014 (at 15:56) by Erwan Jahier> *)


(* let put (os: out_channel) (fmt:('a, unit, string, unit) format4) : 'a = *)
(* 	Printf.kprintf (fun t -> output_string os t) fmt *)

open Printf
open Soc2cIdent
open Data


let rec (type_to_string2 : Data.t -> string) = 
  fun v -> 
    let str =
      match v with
        | Bool -> "bool"
        | Int  -> "int"
        | Real -> "real"
        | Extern s -> id2s s
        | Enum  (s, sl) -> id2s s
        | Struct (sid,_) -> (id2s sid)
        | Array (ty, sz) -> Printf.sprintf "%s_%d" (type_to_string2 ty) sz 
        | Alpha nb -> "alpha_"^(string_of_int nb) 
        | Alias(n,_) -> n
    in
    str

let inlined_soc = Soc2cDep.inlined_soc

(****************************************************************************)

(* Soc printer *)
type 'a soc_pp = {
  hfmt:  ('a, unit, string, unit) format4 -> 'a;
  cfmt:  ('a, unit, string, unit) format4 -> 'a;
  cput : string -> unit;
  hput : string -> unit;
  soc: Soc.t
}

let (string_of_soc_key : Soc.key -> string) =
  fun (name,_,_) ->  (id2s name)

let string_of_flow_decl (id, t) = 
  Printf.sprintf "   %s;\n" (Soc2cUtil.type_to_string t (id2s id)) 


let (is_memory_less : Soc.t -> bool) = SocUtils.is_memory_less

let string_of_var_expr = Soc2cDep.string_of_var_expr
open Soc

(* when an error occur, remove the generated c file (for the nonreg tests) *)
exception Delete_C_files

let var_expr_is_not_a_slice = function Slice _ -> false | _ -> true

let (gao2c : Soc.tbl -> 'a soc_pp -> Soc.gao -> unit) =
  fun stbl sp gao -> 
    let rec gao2str gao = 
      match gao with
        | Case(id, id_gao_l) -> ( 
          let to_case_str (v,gaol) =
            let gaol_str = (List.map gao2str gaol) in
            let gaol_block = String.concat "" gaol_str in
            (id2s v), gaol_block
          in
          let cases = List.map to_case_str id_gao_l in
          let ctx_opt = 
            let il,ol = sp.soc.profile in
            if List.mem_assoc id il || List.mem_assoc id ol then
              (if SocUtils.is_memory_less sp.soc then Soc2cUtil.ML_IO sp.soc.key 
               else Soc2cUtil.M_IO)
            else Soc2cUtil.Local 
          in
          let str = Soc2cUtil.gen_c_switch  (Soc2cDep.ctx_var ctx_opt id) cases in
          str
        )
        | Call(vel_out, Assign, vel_in) -> (
          let l = List.map2 (Soc2cDep.gen_assign_var_expr sp.soc) vel_out vel_in in
          String.concat "" l 
        )
        | Call(vel_out, Method((inst_name,sk),sname), vel_in) -> ( 
          let called_soc = Soc.SocMap.find sk stbl in
          let ctx = Printf.sprintf "ctx->%s" (id2s inst_name) in
          List.iter (fun ve  -> assert(var_expr_is_not_a_slice ve)) vel_in;
          List.iter (fun ve  -> assert(var_expr_is_not_a_slice ve)) vel_out;
          Soc2cDep.gen_step_call sp.soc called_soc vel_out vel_in ctx sname
            ("&ctx->"^(id2s inst_name))
        )
        | Call(vel_out, Procedure sk, vel_in) -> (
          let called_soc = Soc.SocMap.find sk stbl in
          let ctx = get_ctx_name called_soc.key in
          (try
            List.iter (fun ve  -> assert(var_expr_is_not_a_slice ve)) vel_in;
            List.iter (fun ve  -> assert(var_expr_is_not_a_slice ve)) vel_out;
          with _ -> 
            print_string "Error. Slices in left part not yet supported, sorry\n";
            flush stdout;
            raise Delete_C_files
          ); 
          Soc2cDep.gen_step_call sp.soc called_soc vel_out vel_in ctx "step" ""
        )
    in
    sp.cput (gao2str gao)

let (step2c : Soc.tbl -> 'a soc_pp -> Soc.step_method -> unit) =
  fun stbl sp sm -> 
    if inlined_soc sp.soc then () (* don't generate code if inlined *) else
    (*     let sname = Soc2cDep.step_name sp.soc.key sm.name in *)
      let sname = Soc2cDep.step_name sp.soc.key sm.name in
      let ctx = if is_memory_less sp.soc then "" else
          Printf.sprintf "%s_type* ctx" (get_ctx_name sp.soc.key)
      in
      let ctx_decl = if is_memory_less sp.soc then "" else
          Printf.sprintf "%s_type*" (get_ctx_name sp.soc.key)
      in
      if sm.impl<>Extern then (
        sp.hput (Printf.sprintf "void %s(%s);\n" sname ctx_decl);
        sp.cput (Printf.sprintf "void %s(%s){\n" sname ctx);
        (match sm.impl with
          | Extern -> ()
          | Predef -> sp.cput (Soc2cDep.get_predef_op sp.soc.key)
          | Gaol(vl, gaol) -> (
            List.iter (fun v -> sp.cput (string_of_flow_decl v)) vl ; 
            sp.cput "\n"; 
            List.iter (gao2c stbl sp) gaol
          )
          | Iterator(it,it_soc_key,s) ->     
            let it_soc = Soc.SocMap.find it_soc_key stbl in
            sp.cput (Soc2cDep.get_iterator sp.soc it it_soc s)
          | Boolred(i,j,k) -> 
            sp.cput (Soc2cDep.get_boolred sp.soc i j k)
          | Condact(k,el) -> 
            sp.cput (Soc2cDep.get_condact sp.soc (Soc.SocMap.find k stbl) el)
        );
        sp.cput (sprintf "\n} // End of %s\n\n" sname)
      )
let (gen_instance_init_call : 'a soc_pp -> Soc.instance -> unit) =
  fun sp (id,key) -> 
    let ctx_name = get_ctx_name key in
    sp.cfmt "\n    %s_reset(&ctx->%s);" ctx_name (id2s id)

let (soc2c: int -> out_channel -> out_channel -> Soc.tbl -> Soc.t -> unit) = 
  fun pass hfile cfile stbl soc -> 
    if inlined_soc soc then ()  (* don't generate code if inlined *) else
	 let hfmt fmt = Printf.kprintf (fun t -> output_string hfile t) fmt in
	 let cfmt fmt = Printf.kprintf (fun t -> output_string cfile t) fmt in
    let hput str = output_string hfile str in
    let cput str = output_string cfile str in
    let sp = { hfmt = hfmt; cfmt=cfmt; hput = hput; cput = cput; soc = soc } in
    let ctx_name = get_ctx_name soc.key in
    let ctx_name_type = ctx_name^"_type" in    
    if pass=1 then (
       (* Only for ctx of memoryless nodes + main node *)
      if is_memory_less soc then cfmt "%s %s;\n" ctx_name_type ctx_name;
    ) else (
      if is_memory_less soc then () else (
        cfmt "// Memory initialisation for %s\n" ctx_name;
        hfmt "void %s_reset(%s_type* ctx);\n" ctx_name ctx_name;
        cfmt "void %s_reset(%s_type* ctx){" ctx_name ctx_name;
        (* Call the reset_ctx functions of the soc instances *)
        List.iter (gen_instance_init_call sp) soc.instances;
        (match soc.key with
          (* set the parameter fields that have a default value (arrow,fby) *)
          | (_,_,MemInit (ve)) -> 
            assert(var_expr_is_not_a_slice ve);
            cfmt "\n   ctx->_memory = %s;" (string_of_var_expr soc ve)
          | _ -> ()
        );
        cfmt "
}
";      
        cfmt "// Memory allocation for %s\n" ctx_name;
        hfmt "%s_type* %s_new_ctx();\n" ctx_name ctx_name;
        cfmt "%s_type* %s_new_ctx(){" ctx_name ctx_name;
        cfmt "
   %s_type* ctx = (%s_type*)calloc(1, sizeof(%s_type));
   // ctx->client_data = cdata;
   %s_reset(ctx);
   return ctx;
}
" ctx_name ctx_name ctx_name ctx_name) ;
      cfmt "// Step function(s) for %s\n" ctx_name;
      List.iter (step2c stbl sp) soc.step;
      ()
    )

(****************************************************************************)
let (type_to_format_string : Data.t -> string) =
  function
    | Bool -> "%d"
    | Int -> "%d"
    | Real-> "%f"
    | Extern s -> "%s"
    | Enum  (s, sl) -> "%d" 
    | Struct (sid,_) -> "%s"
    | Array (ty, sz) -> "%s"
    | Alpha nb -> assert false 
    | Alias _ -> assert false 


(****************************************************************************)

let (typedef_of_soc : Soc.t -> string) =
  fun soc -> 
    if inlined_soc soc then ""  (* don't generate code if inlined *) else
    let ctx_name = get_ctx_name soc.key in
    let ctx_name_type = ctx_name^"_type" in    
    let il,ol = soc.profile in
    let str = Printf.sprintf  "/* %s */\ntypedef struct {\n   /*INPUTS*/\n" ctx_name in
    let str = List.fold_left (fun acc v -> acc^ (string_of_flow_decl v)) str il in
    let str = str ^ "   /*OUTPUTS*/\n" in
    let str = List.fold_left (fun acc v -> acc^ (string_of_flow_decl v)) str ol in
    let str = str ^
      (match soc.memory with
        | No_mem -> ""
        | Mem t -> 
          Printf.sprintf "   /*Memory cell*/\n   %s ;\n" (id2s (Soc2cUtil.type_to_string t "_memory"))
        | Mem_hidden -> ""
      )
    in
    let str =  str ^ (if soc.instances <> [] then  "   /*INSTANCES*/\n" else "") in
    let string_of_instance (id,sk) = 
      Printf.sprintf "   %s_type %s;\n" (get_ctx_name sk) (id2s id)
    in
    let str =  List.fold_left (fun acc inst -> acc^(string_of_instance inst)) str soc.instances in
    let str = Printf.sprintf  "%s} %s;\n\n" str ctx_name_type in
    str

module KeySet = Set.Make(struct type t = Soc.key let compare = compare end)
module ItemKeySet = Set.Make(struct type t = Lic.item_key let compare = compare end)

(* To perform the topological sort of typedef. nf stands for "no
   fixpoint", that should be done by the caller. it is recursive just
   to deal with array of arrays *)
let  (find_typedep_nf : Lic.type_ -> Lic.item_key list) =
  fun t -> 
    let rec aux top = function
      | Lic.Bool_type_eff | Lic.Int_type_eff | Lic.Real_type_eff
      | Lic.External_type_eff _   | Lic.TypeVar _  
        ->  []
      | Lic.Abstract_type_eff(name,_) 
      | Lic.Enum_type_eff(name,_) -> if top then [] (* avoid self dep *) else [name] 
      | Lic.Array_type_eff(t,_) -> aux false t
      | Lic.Struct_type_eff(name, fl) -> 
        if not top then [name] else List.flatten(List.map (fun (_,(t,_)) -> aux false t) fl)
    in
    aux true t

let (is_extern_type: Lic.type_ -> bool) =
  function 
    | Lic.External_type_eff _ -> true
    | _  -> false

let (typedef : LicPrg.t -> Soc.tbl -> Soc.t -> string) =
  fun licprg soc_tbl main_soc ->
    (* We need to print the ctx typedef a good order
       (w.r.t. typedef dependancies).  To do that, we traverse
       the tree of soc instances which root is the main soc. *)
    let visited = KeySet.empty in
    (* Soc with memory can be used several times; hence we mark via this
       set the ones that have already been visited. *)
    let rec (soc_with_mem : string * KeySet.t -> Soc.t -> string* KeySet.t) =
      fun (acc,visited) soc ->
        if KeySet.mem soc.key visited then (acc,visited) else
          let visited = KeySet.add soc.key visited in
          let acc,visited =
            List.fold_left
              (fun (acc,visited) (iname, sk) ->
                let soc = SocUtils.find_no_exc sk soc_tbl in
                soc_with_mem (acc,visited) soc
              )
              (acc,visited) soc.instances
          in
          let acc = acc ^ (typedef_of_soc soc) in
          acc,visited
    in
    let soc_ctx_typedef_with = 
      if is_memory_less main_soc then "" else fst (soc_with_mem ("",visited) main_soc)
    in
    (* Then we still have to print memoryless soc that can not appear
       as a soc instance *)
    let soc_ctx_typedef_without =
      let socs = Soc.SocMap.bindings soc_tbl in
      let socs = snd (List.split socs) in
      let memless_soc_to_string acc soc =
        if is_memory_less soc then acc^(typedef_of_soc soc) else acc
      in 
      List.fold_left  memless_soc_to_string "" socs
    in 
    let to_c k t =
      Printf.sprintf "typedef %s;\n" (Soc2cUtil.lic_type_to_c t (long2s k))
    in
    let rec (typedef_to_string : Lic.item_key -> Lic.type_ -> string * ItemKeySet.t -> 
             string * ItemKeySet.t) =
      fun k t acc ->
        (* topological sort according to type dep *)
        if is_extern_type t then acc else
        if ItemKeySet.mem k (snd acc) then acc else
          let type_list = find_typedep_nf t in
          let acc = List.fold_left 
            (fun acc k -> 
              match LicPrg.find_type licprg k with
                | Some t -> typedef_to_string k t acc
                | None -> acc (* occurs ? *)
            ) 
            acc type_list 
          in
          ((fst acc)^(to_c k t), ItemKeySet.add k (snd acc))
    in
    let user_typedef,_ = LicPrg.fold_types typedef_to_string licprg ("",ItemKeySet.empty) in

    "// User typedef \n"^user_typedef
    ^"// Memoryless soc ctx typedef \n"^soc_ctx_typedef_without
    ^"// Memoryfull soc ctx typedef \n"^soc_ctx_typedef_with
    



(****************************************************************************)
let rec (const_to_c: Lic.const -> string) =
  function
    | Lic.Bool_const_eff true -> "1"
    | Lic.Bool_const_eff false -> "0"
    | Lic.Int_const_eff i -> (sprintf "%s" i)
    | Lic.Real_const_eff r -> r
    | Lic.Extern_const_eff (s,t) -> (long2s s)
    | Lic.Abstract_const_eff (s,t,v,_) -> const_to_c v
    | Lic.Enum_const_eff   (s,Lic.Enum_type_eff(_,ll)) -> Lic.enum_to_string s ll
    | Lic.Enum_const_eff   _ -> assert false (* SNO *)
    | Lic.Struct_const_eff (fl, t) -> (
      let string_of_field = 
        function (id, veff) -> 
          (Ident.to_string id)^" = "^ (const_to_c veff) 
      in
      let flst = List.map string_of_field fl in
(*       (string_of_type_eff t)^ *)
        "{"^(String.concat "; " flst)^"}"
    )
    | Lic.Array_const_eff (ctab, t) -> (
      let vl = List.map const_to_c ctab in
      "["^(String.concat ", " vl)^"]"
    )
    | Lic.Tuple_const_eff   cl -> assert false

let (constdef : LicPrg.t -> string) = 
  fun licprg -> 
    let to_c k = function
    | Lic.Extern_const_eff _ -> ""
    | c  -> 
      Printf.sprintf "#define %s %s\n"
(*       Printf.sprintf "const %s = %s;\n" *)
        (long2s k)
        (const_to_c c)
    in
    LicPrg.fold_consts (fun k t acc -> acc ^ (to_c k t)) licprg  "\n// Constant definitions \n"


(****************************************************************************)

let (gen_memoryless_ctx : Soc.tbl -> string) =
  fun stbl -> 
    let acc = Printf.sprintf "\n// Allocation of memoryless ctx\n" in
    let do_soc sk soc acc =
      if is_memory_less soc && not (inlined_soc soc) then 
        let ctx_name = get_ctx_name soc.key in
        let ctx_name_type = ctx_name^"_type" in    
        Printf.sprintf "%sextern %s %s;\n" acc ctx_name_type ctx_name 
      else acc
    in
    Soc.SocMap.fold do_soc stbl acc
      
(****************************************************************************)
let (gen_loop_file : Soc.t -> out_channel -> Soc.tbl -> unit) =
  fun soc oc stbl -> 
    let base = (string_of_soc_key soc.key) in
    let putc s =  output_string oc s in
    let ctx = get_ctx_name soc.key in
    let step = Soc2cDep.step_name soc.key "step" in
    let (n,_,_) = soc.key in
    let n = id2s n in
    let inputs,outputs = soc.profile in
    let inputs_io = SocVar.expand_profile true false inputs in
    let outputs_io = SocVar.expand_profile true false  outputs in
    let inputs = SocVar.expand_profile true true inputs in
    let outputs = SocVar.expand_profile true true  outputs in
    Lv6util.entete oc "/*" "*/";
    putc  ("
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include \""^base ^".h\" 
/* Print a promt ? ************************/
static int ISATTY;
/* MACROS DEFINITIONS ****************/
#ifndef TT
#define TT \"1\"
#endif
#ifndef FF
#define FF \"0\"
#endif
#ifndef BB
#define BB \"bottom\"
#endif
#ifdef CKCHECK
/* set this macro for testing output clocks */
#endif

/* Standard Input procedures **************/
_boolean _get_bool(char* n){
   char b[512];
   _boolean r = 0;
   int s = 1;
   char c;
   do {
      if(ISATTY) {
         if((s != 1)||(r == -1)) printf(\"\\a\");
         // printf(\"%s (1,t,T/0,f,F) ? \", n);
      }
      if(scanf(\"%s\", b)==EOF) exit(0);
      if (*b == 'q') exit(0);
      s = sscanf(b, \"%c\", &c);
      r = -1;
      if((c == '0') || (c == 'f') || (c == 'F')) r = 0;
      if((c == '1') || (c == 't') || (c == 'T')) r = 1;
   } while((s != 1) || (r == -1));
   return r;
}
_integer _get_int(char* n){
   char b[512];
   _integer r;
   int s = 1;
   do {
      if(ISATTY) {
         if(s != 1) printf(\"\\a\");
         //printf(\"%s (integer) ? \", n);
      }
      if(scanf(\"%s\", b)==EOF) exit(0);
      if (*b == 'q') exit(0);
      s = sscanf(b, \"%d\", &r);
   } while(s != 1);
   return r;
}
#define REALFORMAT ((sizeof(_real)==8)?\"%lf\":\"%f\")
_real _get_real(char* n){
   char b[512];
   _real r;
   int s = 1;
   do {
      if(ISATTY) {
         if(s != 1) printf(\"\\a\");
         //printf(\"%s (real) ? \", n);
      }
      if(scanf(\"%s\", b)==EOF) exit(0);
      if (*b == 'q') exit(0);
      s = sscanf(b, REALFORMAT, &r);
   } while(s != 1);
   return r;
}
/* Standard Output procedures **************/
void _put_bottom(char* n){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%s \", BB);
   if(ISATTY) printf(\"\\n\");
}
void _put_bool(char* n, _boolean _V){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%s \", (_V)? TT : FF);
   if(ISATTY) printf(\"\\n\");
}
void _put_int(char* n, _integer _V){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%d \", _V);
   if(ISATTY) printf(\"\\n\");
}
void _put_real(char* n, _real _V){
   if(ISATTY) printf(\"%s = \", n);
   printf(\"%f \", _V);
   if(ISATTY) printf(\"\\n\");
}
/* Output procedures **********************/
#ifdef CKCHECK
void %s_BOT_n(void* cdata){
   _put_bottom(\"n\");
}
#endif
/* Output procedures **********************/
void "^n^"_O_n(void* cdata, _integer _V) {
   _put_int(\"n\", _V);
}"^ (gen_memoryless_ctx stbl) ^
"/* Main procedure *************************/
int main(){
   
   int s = 0;
   /* Context allocation */
   "^ctx^"_type* ctx = "^(if is_memory_less soc then (
     "&"^ ctx^ ";"
   ) else (ctx^"_new_ctx(NULL);")));

    let to_rif_decl (n,t) = ("\\\""^n^"\\\":" ^(type_to_string t)) in
    let inputs_t  = List.map to_rif_decl inputs_io in
    let outputs_t = List.map to_rif_decl outputs_io in
    let inputs_decl = Printf.sprintf "#inputs %s" (String.concat " " inputs_t) in
    let outputs_decl = Printf.sprintf "#outputs %s" (String.concat " " outputs_t) in
    
    putc  ("
    printf(\""^inputs_decl^"\\n\");
    printf(\""^outputs_decl^"\\n\");

   /* Main loop */
   ISATTY = isatty(0);
   while(1){
      if (ISATTY) printf(\"#step %d \\n\", s+1);
      else if(s) printf(\"\\n\");
      fflush(stdout);
      ++s;
");
    List.iter (fun (id,t) -> 
      let t = type_to_string2 t in
      let str = Printf.sprintf "      ctx->%s = _get_%s(\"%s\");\n" id t id in
      putc str
    ) 
      inputs;
    let inputs_fmt  = List.map (fun (_,t) -> type_to_format_string t) inputs in
    let outputs_fmt = List.map (fun (_,t) -> type_to_format_string t) outputs in
    putc ("      " ^ step^"(ctx);
      // printf(\"" ^ (String.concat " " inputs_fmt)^ " #outs " ^ 
             (String.concat " " outputs_fmt)^ "\\n\"," ^
             (String.concat "," (List.map (fun (id,_) -> "ctx->"^id ) (inputs@outputs)))^ 
            ");
      printf(\"" ^ 
             (String.concat " " outputs_fmt)^ "\\n\"," ^
             (String.concat "," (List.map (fun (id,_) -> "ctx->"^id ) (outputs)))^ 
            ");
   }
   return 1;
   
}
")  

(****************************************************************************)
(* The entry point for lus2lic --to-c *)
let (f : Lv6MainArgs.t -> Soc.key -> Soc.tbl -> LicPrg.t -> unit) =
  fun args msoc stbl licprg -> 
    let socs = Soc.SocMap.bindings stbl in
    let socs = snd (List.split socs) in 
    (* XXX que fait-on pour les soc predef ? *)
    (*     let _, socs = List.partition is_predef socs in *)
    let base = string_of_soc_key msoc in
    let hfile = base ^  ".h" in
    let cfile = base ^ ".c" in
    let ext_cfile = Printf.sprintf "%s_ext.c" base in
    let ext_hfile = Printf.sprintf "%s_ext.h" base in
    let loopfile = base^"_loop.c" in
    let occ = open_out cfile in
    let och = open_out hfile in
    let ocl = open_out loopfile in
    try
      let putc s =  output_string occ s in
      let puth s =  output_string och s in
      let main_soc = Soc.SocMap.find msoc stbl in
    (* Generate ext files if necessary  *)
      let needs_cfile, needs_hfile = 
        Soc2cExtern.gen_files main_soc stbl licprg ext_cfile ext_hfile hfile
      in

      Lv6util.entete occ "/*" "*/" ;
      Lv6util.entete och "/*" "*/";
      gen_loop_file main_soc ocl stbl;
      
      output_string och "
#include <stdlib.h>
#include <string.h>

#ifndef _SOC2C_PREDEF_TYPES
#define _SOC2C_PREDEF_TYPES
typedef int _boolean;
typedef int _integer;
typedef char* _string;
typedef double _real;
typedef double _double;
typedef float _float;
#define _false 0
#define _true 1
#endif
// end of _SOC2C_PREDEF_TYPES

";

    if needs_hfile then puth (Printf.sprintf "#include \"%s\"\n" ext_hfile);
    puth (Printf.sprintf "#ifndef _%s_LUS2LIC_H_HILE\n" base);
    puth (Printf.sprintf "#define _%s_LUS2LIC_H_HILE\n" base);

    putc (Printf.sprintf "#include \"%s\"\n" hfile);
    puth (typedef licprg stbl main_soc );
    putc (constdef licprg);
(*     putc (Soc2cExtern.cpy_declaration licprg); *)
    putc (Soc2cExtern.const_declaration licprg);
    puth "/////////////////////////////////////////////////\n";
    puth "//// Static allocation of memoryless soc ctx\n";
    List.iter (soc2c 1 och occ stbl) socs;
    puth "/////////////////////////////////////////////////\n";
    putc "//// Defining step functions\n";
    List.iter (soc2c 2 och occ stbl) socs;
    
    puth "/////////////////////////////////////////////////\n";
    let l =  (Soc2cGenAssign.gen_used_types socs) in 
    if l <> [] then puth "// Defining array and extern types assignments \n";
    List.iter (fun t -> puth (Soc2cGenAssign.f t)) l;
    
    puth "#endif\n";
    flush occ; close_out occ;
    flush och; close_out och;
    flush ocl; close_out ocl;
    Printf.printf "%s has been generated.\n" loopfile;
    Printf.printf "%s has been generated.\n" hfile; 
    Printf.printf "%s has been generated.\n" cfile;

    let execfile = if args.outfile = "" then "a.out" else args.outfile in
    let gcc =
      if needs_cfile then 
        Printf.sprintf " gcc -o %s %s %s %s" execfile cfile ext_cfile loopfile 
      else 
        Printf.sprintf "gcc  -o %s %s %s"  execfile cfile loopfile 
    in
    if args.Lv6MainArgs.launch_cc then (
      if (Sys.command gcc)=0 then 
        print_string ("sys call: '"^gcc^"' executed with success\n")
      else
        print_string ("sys call: '"^gcc^"' failed\n")
    ) else
      print_string ("you can compile those files doing, e.g.,\n  "^gcc^"\n");
    flush stdout
    with
      | Delete_C_files -> 
        close_out occ;
        close_out och;
        close_out ocl;
        if Sys.file_exists hfile then Sys.remove hfile;
        if Sys.file_exists cfile then Sys.remove cfile;
        if Sys.file_exists ext_cfile then Sys.remove ext_cfile;
        if Sys.file_exists ext_hfile then Sys.remove ext_hfile;
        if Sys.file_exists loopfile then Sys.remove loopfile;
        exit 2
