(* Time-stamp: <modified the 08/10/2014 (at 18:23) by Erwan Jahier> *)

open Data
open Lic
open Lxm

module OrderedData = struct
  type t = Data.t
  let compare = compare
end

module DataSet = Set.Make(OrderedData)

let rec (update_set : DataSet.t -> Data.t -> DataSet.t) =
fun acc t -> 
  match t with 
    | Data.Array(st,_) -> update_set (DataSet.add t acc) st
    | Data.Extern _ -> DataSet.add t acc 
    | Struct(_,fl)  -> List.fold_left (fun acc (_,t) -> update_set acc t) acc fl
    | Alias(n,t) -> update_set acc t
    | _ -> acc

open Soc

let (update_set_soc_var : DataSet.t -> Soc.var -> DataSet.t) =
  fun acc (_,datat) -> 
    update_set acc datat

let (update_set_step:  DataSet.t -> Soc.step_method -> DataSet.t) =
  fun acc sm -> 
    match sm.impl with
      | Gaol(vl,_) -> List.fold_left update_set_soc_var acc vl
      | _ -> acc


(* exported *)
let (gen_used_types : Soc.t list -> Data.t list) =
  fun socs -> 
    let rec aux acc soc =
      let ins, outs = soc.profile in
      let acc = List.fold_left update_set_soc_var acc ins in
      let acc = List.fold_left update_set_soc_var acc outs in
      let acc = List.fold_left update_set_step acc soc.step in
      acc
    in
    let acc = List.fold_left aux DataSet.empty socs in
    DataSet.elements acc
      

(* exported *)
let (f: Data.t -> string) =
  fun t -> 
    let t_str = Soc2cIdent.type_to_short_string t in
    let t_def = Printf.sprintf "
#ifndef _assign_%s 
#define _assign_%s(dest, source, size) memcpy(dest, source, size)
#endif
"
      t_str t_str
    in
    t_def


