(* Time-stamp: <modified the 07/10/2014 (at 16:35) by Erwan Jahier> *)
open Soc

let colcol = Str.regexp "::"
let id2s id = (* XXX Refuser les noms de module à la con plutot *)
  let str =
	 match Str.split colcol id with
	   | ["true"] -> "_true" (* a good idea to do that here ? *)  
	   | ["false"] -> "_false"
	   | [s] -> s
	   | [m;s] -> if Lv6MainArgs.global_opt.Lv6MainArgs.no_prefix then s else m^"_"^s 
	   | _ -> id
  in
  let str = Str.global_replace colcol "_" str in
  str

let long2s l = id2s (Ident.string_of_long l)

let rec (type_to_short_string : Data.t -> string) = 
  fun v -> 
    let str =
      match v with
        | Data.Bool -> "b"
        | Data.Int -> "i"
        | Data.Real-> "r"
        | Data.Extern s -> id2s s 
        | Data.Enum  (s, sl) -> "e" (* s *) 
        | Data.Struct (sid,_) -> id2s sid
        | Data.Array (ty, sz) -> Printf.sprintf "%sp%d" (type_to_short_string ty) sz 
        | Data.Alias(n,_) -> n
        | Data.Alpha nb ->
        (* On génère des "types" à la Caml : 'a, 'b, 'c, etc. *)
          let a_value = Char.code('a') in
          let z_value = Char.code('z') in
          let str =
            if (nb >= 0 && nb <= (z_value - a_value)) then
              ((Char.escaped (Char.chr(a_value + nb))))
            else
              ("a" ^ (string_of_int nb))
          in
          str
    in
    str

let (key_op2str : Soc.key_opt -> string) = function
  | Nomore -> ""
  | Curr(cc) -> Ident.string_of_long2 cc
  | Slic(b,e,s) -> Printf.sprintf "_slice_%d_%d_%d" b e s
  | MemInit(var_expr) ->  "_" ^
    (* XXX This is wrong because hash is not an injection !!! *)
    (string_of_int (Hashtbl.hash var_expr))


let (get_base_name : Soc.key -> string) =
  fun (name,tl,opt) -> 
    let l = List.map type_to_short_string tl in
    (id2s (Printf.sprintf "%s_%s%s" (id2s name) (String.concat "" l) (key_op2str opt))) 

let (get_ctx_name : Soc.key -> string) =
  fun sk -> (get_base_name sk) ^ "_ctx"

let (get_soc_name : Soc.key -> string) =
  fun sk -> (get_base_name sk)
