(* Time-stamp: <modified the 08/10/2014 (at 10:59) by Erwan Jahier> *)

(* A local exception used to check if a predef is supported. 
   The idea is that when gen_call_do is called  with empty lists,
   it means that it is called to check if the predef is supported.

   This (ugly but local) trick is to avoid to duplicate the match
   done in gen_call_do 
*)
exception Is_supported

let stdbin op ll rl = 
	match (ll,rl) with
	| ([l], [r1;r2]) -> Printf.sprintf "  %s = %s %s %s;\n" l r1 op r2
	| ([],[]) -> raise Is_supported
	| _ -> assert false

let stduna op ll rl = 
	match (ll,rl) with
	| ([l], [r]) -> Printf.sprintf "  %s = %s %s;\n" l op r
	| ([],[]) -> raise Is_supported
	| _ -> assert false

let stdimpl ll rl = 
	match (ll,rl) with
	| ([l], [r1;r2]) -> Printf.sprintf   "%s = !%s | %s;\n" l r1 r2
	| ([],[]) -> raise Is_supported
	| _ -> assert false


(* exported *)
(* ZZZ code dupl with SocPredef2cHeap.get_predef_op *)
let (gen_call_do :  string -> Soc.t -> string list -> string list -> string) =
  fun op soc vel_in vel_out -> 
    let lstduna str = stduna str vel_in vel_out in
    let lstdbin str = stdbin str vel_in vel_out in
    let lstdimpl () = stdimpl vel_in vel_out in
    match op with
	   | "Lustre::uminus" -> lstduna "-"
	   | "Lustre::iuminus" -> lstduna "-"
	   | "Lustre::ruminus" -> lstduna "-"
	   | "Lustre::not"     -> lstduna "!"

	   | "Lustre::mod"    -> lstdbin "%" 
	   | "Lustre::plus"   -> lstdbin "+" 
	   | "Lustre::iplus"  -> lstdbin "+" 
	   | "Lustre::rplus"  -> lstdbin "+" 
	   | "Lustre::times"  -> lstdbin "*"
	   | "Lustre::itimes" -> lstdbin "*"
	   | "Lustre::rtimes" -> lstdbin "*"
	   | "Lustre::div"    -> lstdbin "/"
	   | "Lustre::idiv"   -> lstdbin "/"
	   | "Lustre::rdiv"   -> lstdbin "/"
	   | "Lustre::minus"  -> lstdbin "-"
	   | "Lustre::iminus" -> lstdbin "-"
	   | "Lustre::rminus" -> lstdbin "-"

	   | "Lustre::real2int" -> lstduna "(_integer)"
	   | "Lustre::int2real" -> lstduna "(_real)"

	   | "Lustre::lt"  -> lstdbin "<"
	   | "Lustre::ilt" -> lstdbin "<"
	   | "Lustre::lt"  -> lstdbin "<"
	   | "Lustre::rlt" -> lstdbin "<"
	   | "Lustre::gt"  -> lstdbin ">"
	   | "Lustre::igt" -> lstdbin ">"
	   | "Lustre::rgt" -> lstdbin ">"
	   | "Lustre::lte" -> lstdbin "<="
	   | "Lustre::ilte"-> lstdbin "<="
	   | "Lustre::rlte"-> lstdbin "<="
	   | "Lustre::gte" -> lstdbin ">="
	   | "Lustre::igte"-> lstdbin ">="
	   | "Lustre::rgte"-> lstdbin ">="

	   | "Lustre::and" -> lstdbin "&"
	   | "Lustre::or"  -> lstdbin "|"
	   | "Lustre::xor" -> lstdbin "!="
	   | "Lustre::impl" -> lstdimpl ()

	   | "Lustre::equal" -> lstdbin "=="
	   | "Lustre::diff" -> lstdbin "!="

	   | _ -> raise Not_found

let (gen_call :  string -> Soc.t -> string list -> string list -> string) =
  fun op soc vel_in vel_out -> 
    assert ((vel_in,vel_out) <> ([],[]));
    gen_call_do op soc vel_in vel_out

let (is_call_supported : string -> Soc.t -> bool) =
  fun op soc -> 
    try ignore (gen_call_do op soc [] []); assert false (* sno *)
    with 
      | Is_supported -> true
      | Not_found -> false  
